# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.


'''
Module functional_basis.

'''

from abc import abstractmethod
from scipy.sparse import diags
import numpy as np
import tensap


class FunctionalBasis:
    '''
    Class FunctionalBasis.

    Attributes
    ----------
    measure : tensap.Measure
        The measure associated with the FunctionalBasis.

    is_orthonormal : bool
        Indicates if the basis is orthonormal with respect to the associated
        measure.

    '''

    def __init__(self):
        '''
        Constructor for the class FunctionalBasis.

        Returns
        -------
        None.

        '''
        self.measure = None
        self.is_orthonormal = False

    def __repr__(self):
        return ('<{}:{n}' +
                '{t}measure = {},{n}' +
                '{t}is_orthonormal = {}>').format(self.__class__.__name__,
                                                  self.measure,
                                                  self.is_orthonormal,
                                                  t='\t', n='\n')

    def random(self, n=1, measure=None):
        '''
        Evaluate the basis using n points drawn randomly according to measure
        if provided, or to self.measure otherwise.

        Parameters
        ----------
        n : int, optional
            The number of random evaluations. The default is 1.
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used for the generation of the input
            points. The default is None, indicating to use self.measure.

        Returns
        -------
        basis_eval : numpy.ndarray
            Random evaluations of the basis functions.
        x : numpy.ndarray
            The input points.

        '''
        if measure is None:
            measure = self.measure
        if not isinstance(measure, tensap.ProbabilityMeasure):
            raise ValueError('Must provide a ProbabilityMeasure.')

        x = measure.random(n)
        basis_eval = np.reshape(self.eval(x), (n, self.cardinal()))
        return basis_eval, x

    @staticmethod
    def storage():
        '''
        Return the storage requirement of the FunctionalBasis.

        Returns
        -------
        int
            The storage requirement of the FunctionalBasis.

        '''
        return 0

    def adaptation_path(self):
        '''
        Return the adaptation path of the functional basis.

        Returns
        -------
        numpy.ndarray
            Boolean array, where n is the dimension of the functional basis,
            and m is the number of elements in the adaptation path,
            column P[:,i] corresponds to a sparsity pattern.

        '''
        return np.triu(np.full([self.cardinal()]*2, True))

    def interpolate(self, y, x=None):
        '''
        Provide an interpolation on a functional basis of a function (or values
        of the function) y associated with a set of n interpolation points x.

        Parameters
        ----------
        y : function or list or numpy.ndarray
            The function to interpolate, or values of it.
        x : list or numpy.ndarray, optional
            The interpolation points. The default is None, indicating to
            deduce them from the basis.

        Returns
        -------
        f : tensap.FunctionalBasisArray
            The computed interpolation.

        '''
        if x is None:
            x = self.interpolation_points()
        try:
            y = y(x)
        except Exception:
            pass

        if np.ndim(y) == 1:
            y = np.reshape(y, [-1, 1])

        hx = self.eval(x)
        data = np.linalg.solve(hx, y)
        f = tensap.FunctionalBasisArray(data, self, np.shape(y)[1])
        f.measure = self.measure
        return f

    @staticmethod
    def mean():
        '''
        Return the mean of the basis functions.

        Returns
        -------
        numpy.ndarray
            The mean of the basis functions.

        '''
        raise NotImplementedError('No generic implementation of the method.')

    def expectation(self):
        '''
        Return the expectation of the basis functions. Equivalent to
        self.mean().

        Returns
        -------
        numpy.ndarray
            The expectation of the basis functions.

        '''
        return self.mean()

    @staticmethod
    def conditional_expectation():
        '''
        Compute the conditional expectation of self with respect to
        the random variables dims (a subset of range(d)). The expectation
        with respect to other variables (in the complementary set of
        dims) is taken with respect to the probability measure given by
        tensap.RandomVector XdimsC if provided, or with respect to the
        probability measure associated with the corresponding bases of
        the function.

        Parameters
        ----------
        dims : numpy.ndarray
            The dimensions in which the expectation is computed.
        XdimsC: tensap.RandomVector, optional
            The random vector used for the computation of the conditional
            expectation. The default is None, indicating to use the
        probability measure associated with the basis.

        Returns
        -------
        f : tensap.FunctionalBasisArray
            The conditional expectation of the function.

        '''
        raise NotImplementedError('No generic implementation of the method.')

    @staticmethod
    def kron():
        '''
        Compute the Kronecker product of two bases. For two functional bases
        f_i, i = 1, ..., n and g_j, j = 1, ..., m, return a functional basis
        h_k, k = 1, ..., nm.

        Returns
        -------
        tensap.FunctionalBasis
            The obtained basis.

        '''
        NotImplementedError('Method not implemented.')

    def projection(self, fun, G):
        '''
        Compute the projection of the function fun onto the functional basis
        using the integration rule G.

        Parameters
        ----------
        fun : function or tensap.Function
            The function to project.
        G : tensap.IntegrationRule
            The integration rule used for the projection.

        Returns
        -------
        tensap.FunctionalBasisArray
            The projection of the function fun onto the functional basis using
            the integration rule G.

        '''

        A = self.eval(G.points)
        W = diags(G.weights)

        y = fun(G.points)
        if self.is_orthonormal:
            u = np.matmul(np.transpose(A), W.dot(y))
        else:
            u = np.linalg.solve(np.matmul(np.transpose(A), W.dot(A)),
                                np.matmul(np.transpose(A), W.dot(y)))
        if u.ndim == 1:
            u = np.reshape(u, [-1, 1])
        return tensap.FunctionalBasisArray(u, self, u.shape[1])

    def interpolation_points(self, *args):
        '''
        Return the interpolation points for the basis.

        See also FunctionalBasis.magic_points.

        Parameters
        ----------
        *args : tuple
            The inputs arguments of the method FunctionalBasis.magic_points.

        Returns
        -------
        numpy.ndarray
            The interpolation points for the basis.

        '''
        return self.magic_points(*args)[0]

    def magic_points(self, x=None, J=None):
        '''
        Provide the magic points associated with a functional basis f selected
        in a given set of points x.

        The method uses magicIndices(F,numel(f)) on the matrix F of evaluations
        of f at points x.

        Parameters
        ----------
        x : list or numpy.ndarray, optional
            The points used to construct the matrix F. The default is None,
            indicating to choose x automatically based on self.measure.
        J : numpy.ndarray, optional
            The default is None. If not none, selected the magic indices with
            tensap.magic_indices(F[:, J], self.cardinal(), 'left')[0]

        Returns
        -------
        points : numpy.ndarray
            The magic points.
        ind : numpy.ndarray
            The locations of the magic points in x.
        output : dict
            A dictionnary of outputs of the method.

        '''
        if x is None:
            if isinstance(self.measure, (tensap.DiscreteMeasure,
                                         tensap.DiscreteRandomVariable)):
                x = self.measure.values
            else:
                x = self.measure.random(self.cardinal()*100)

        if np.ndim(x) == 1:
            x = np.expand_dims(x, 1)

        assert x.shape[0] >= self.cardinal(), \
            ('The number of points must be higher than the number of basis ' +
             'functions.')

        F = self.eval(x)
        if J is not None:
            ind = tensap.magic_indices(F[:, J], self.cardinal(), 'left')[0]
        else:
            ind = tensap.magic_indices(F)[0]
        points = x[ind, :]
        if np.ndim(points) == 1:
            points = np.expand_dims(points, 1)

        # Estimation of the Lebesgue constant
        h_x = np.transpose(self.eval(points))
        A = np.transpose(np.linalg.solve(h_x, np.transpose(F)))

        output = {'lebesgue_constant': np.max(np.sum(np.abs(A), 1))}

        return points, ind, output

    def domain(self):
        '''
        Return the domain of the set of basis functions, which is the support
        of the associated measure.

        Returns
        -------
        numpy.ndarray
            The domain of the set of basis functions

        '''
        return self.measure.support()

    def christoffel(self, x):
        # TODO christoffel
        raise NotImplementedError('Method not implemented.')

    def orthonormalize(self):
        '''
        Orthonormalize the basis.

        Returns
        -------
        out : tensap.SubFunctionalBasis
            The orthonormalized basis.

        '''
        G = self.gram_matrix()
        out = self
        if np.linalg.norm(G - np.eye(G.shape[0]), 2):
            A = np.linalg.inv(np.linalg.cholesky(G))
            out = tensap.SubFunctionalBasis(self, np.transpose(A))
        out.is_orthonormal = True
        return out

    def optimal_sampling_measure(self):
        # TODO optimal_sampling_measure
        raise NotImplementedError('Method not implemented.')

    def plot(self, indices=None, n=10000, *args):
        '''
        Plot the functions of the basis.

        Parameters
        ----------
        indices : list or numpy.ndarray, optional
            Indices of the functions to be plotted. The default is None,
            indicating all the functions.
        n : int, optional
            The number of points used for the plot. The default is 10000.
        *args : tuple
            Additional parameters used by matplotlib.pyplot's function plot.

        Returns
        -------
        None.

        '''
        assert self.ndim() == 1, 'Method not implemented.'

        import matplotlib.pyplot as plt

        sup = self.measure.truncated_support()
        if np.size(n) == 1:
            x = np.linspace(sup[0], sup[1], n)
        else:
            x = np.ravel(n)

        if indices is None:
            hx = self.eval(x)
        else:
            hx = self.eval(x, indices)

        plt.plot(x, hx, *args)
        plt.show()

    @abstractmethod
    def cardinal(self):
        '''
        Return the number of basis functions.

        Returns
        -------
        int
            The number of basis functions.

        '''

    @abstractmethod
    def ndim(self):
        '''
        Return the dimension n for f defined in R^n.

        Returns
        -------
        int
            The dimension n for f defined in R^n.

        '''

    @abstractmethod
    def eval(self, x):
        '''
        Return the evaluation of the basis functions at the points x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the basis functions are to be evaluted.

        Returns
        -------
        numpy.ndarray
            The evaluations of the basis functions at the points x.

        '''



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on FullTensor.

'''

import numpy as np
import matplotlib.pyplot as plt
import tensap

# %% Creation of random FullTensor with precribed shapes
# Order-4 full tensor with i.i.d. entries drawn according to the uniform
# distribution on (0, 1)

TENSOR_1 = tensap.FullTensor.rand([2, 3, 4, 5])
print('TENSOR_1 = tensap.FullTensor.rand([2, 3, 4, 5]) =')
print(TENSOR_1)

# Order-6 full tensor with i.i.d. entries drawn according to the standard
# gaussian distribution
TENSOR_2 = tensap.FullTensor.randn([2, 3, 4, 4, 3, 2])
print('\nTENSOR_2 = tensap.FullTensor.rand([2, 3, 4, 4, 3, 2]) =')
print(TENSOR_2)

print('\nNorm of TENSOR_1 = %f' % TENSOR_1.norm())
print('Number of entries of TENSOR_1 = %i' % TENSOR_1.storage())
print('Number of non-zero entries of TENSOR_1 = %i' %
      TENSOR_1.sparse_storage())

# %% Operations on FullTensor
# Contraction of the tensors TENSOR_1 and TENSOR_2, in the dimensions 0 and 2,
# and 5 and 3, respectively
# The resulting tensor TENSOR_3 is of order TENSOR_3.order = 4+6-4 = 6, and of
# dimensions TENSOR_3.shape = [3, 5, 2, 3, 4, 3]

TENSOR_3 = TENSOR_1.tensordot(TENSOR_2, [0, 2], [5, 3])
print('\nTENSOR_3 = TENSOR_1.tensordot(TENSOR_2, [0, 2], [5, 3]) =')
print(TENSOR_3)

# Contraction of the tensors TENSOR_1 and TENSOR_2, in the dimensions 0 and 2,
# and 5 and 3, respectively, and evaluation of the diagonal in the dimensions
# 1 and 4, respectively
# The resulting tensor TENSOR_4 is of order TENSOR_4.order = 4+6-4-1 = 5, and
# of dimensions TENSOR_4.shape = [3, 5, 2, 3, 4]
TENSOR_4 = TENSOR_1.tensordot_eval_diag(TENSOR_2, [0, 2], [5, 3], 1, 4)
print('\nTENSOR_4 = TENSOR_1.tensordot_eval_diag(TENSOR_2, [0, 2], ' +
      '[5, 3], 1, 4) =')
print(TENSOR_4)

# Outer product of the tensors TENSOR_1 and TENSOR_2 with evaluation of the
# diagonal in the dimensions 2 and 3, respectively
# The resulting tensot TENSOR_5 is such that
# TENSOR_5[i1, i2, k, i4, j1, j2, j3, j5, j6] = TENSOR_1[i1, i2, k, i4]
# TENSOR_2(j1, j2, j3, k, j5, j6)
TENSOR_5 = TENSOR_1.outer_product_eval_diag(TENSOR_2, 2, 3)
print('\nTENSOR_5 = TENSOR_1.outer_product_eval_diag(TENSOR_2, 2, 3) =')
print(TENSOR_5)

# Outer product of the tensors TENSOR_1 and TENSOR_2 with evaluation of several
# diagonals, in the dimensions 1 and 2, and 4 and 3 respectively
# The resulting tensor TENSOR_6 is such that
# TENSOR_6(i1, k, l, i4, j1, j2, j3, j6) = TENSOR_1(i1, k, l, i4)
# TENSOR_2(j1, j2, j3, l, k, j6)
TENSOR_6 = TENSOR_1.outer_product_eval_diag(TENSOR_2, [1, 2], [4, 3], True)
print('\nTENSOR_6 = TENSOR_1.outer_product_eval_diag(TENSOR_2, [1, 2], ' +
      '[4, 3], True) =')
print(TENSOR_6)

# %% Orthogonalization of a FullTensor
# The tensor TENSOR_7 is of same shape and order at TENSOR_4, but is such that
# its  {2}-matricization, denoted by MAT_TENSOR_7, verifies
# np.matmul(MAT_TENSOR_7, np.transpose(MAT_TENSOR_7)) =
# np.eye(TENSOR_4.shape[2])
TENSOR_7 = TENSOR_4.orth(2)[0]
print('\nTENSOR_7 = TENSOR_4.orth(2) =')
print(TENSOR_7)

# We check the orthogonality by computing the product of the {2}-matricization
# of TENSOR_7 by itself
MAT_TENSOR_7 = TENSOR_7.matricize(2)
P = MAT_TENSOR_7.tensordot(MAT_TENSOR_7, 1, 1)
print('\nChecking the orthogonality condition on TENSOR_7:')
print(P.data)

# %% Singular values of a FullTensor
SIN_VAL = TENSOR_3.singular_values()

ORDER = TENSOR_3.order
SIZE_2 = int(np.floor(np.sqrt(ORDER)))
SIZE_1 = int(np.ceil(ORDER/SIZE_2))
for nb_plot in np.arange(1, ORDER+1):
    plt.subplot(SIZE_1, SIZE_2, nb_plot)
    plt.bar(range(SIN_VAL[nb_plot-1].size), SIN_VAL[nb_plot-1])
    plt.title('%i-singular values' % (nb_plot-1))
plt.show()



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.


'''
Module functional_basis_array.

'''

from copy import deepcopy
import numpy as np
import tensap


class FunctionalBasisArray(tensap.Function):
    '''
    Class FunctionalBasisArray.

    Attributes
    ----------
    data : numpy.ndarray, optional
            The coefficents of the function on the basis. The default is None.
    basis : tensap.FunctionalBasis, optional
        The basis. The default is None.
    shape : list or numpy.ndarray, optional
        Array such that the function is with values in
        R^(shape[0] x shape[1] x ...). The default is 1.

    '''

    def __init__(self, data=None, basis=None, shape=None):
        '''
        Constructor for the class FunctionalBasisArray

        Parameters
        ----------
        data : numpy.ndarray, optional
            The coefficents of the function on the basis. The default is None.
        basis : tensap.FunctionalBasis, optional
            The basis. The default is None.
        shape : list or numpy.ndarray, optional
            Array such that the function is with values in
            R^(shape[0] x shape[1] x ...). The default is 1.

        Returns
        -------
        None.

        '''
        tensap.Function.__init__(self)

        if shape is None and (data is not None or basis is not None):
            shape = 1

        self.data = np.ravel(data)
        self.basis = deepcopy(basis)
        self.shape = np.atleast_1d(shape)
        self.output_shape = self.shape
        self.data = np.reshape(self.data,
                               np.concatenate(([basis.cardinal()],
                                               self.shape)))

    def __add__(self, g):
        return FunctionalBasisArray(self.data + g.data, self.basis, self.shape)

    def __neg__(self):
        return FunctionalBasisArray(-self.data, self.basis, self.shape)

    def __sub__(self, g):
        return FunctionalBasisArray(self.data - g.data, self.basis, self.shape)

    def __mul__(self, g):
        if isinstance(g, FunctionalBasisArray):
            g = g.data
        return FunctionalBasisArray(self.data * g, self.basis, self.shape)

    def matmul(self, v):
        '''
        Compute the matrix multiplication of self.data with v.

        Parameters
        ----------
        v : numpy.ndarray
            The array used in the matrix multiplication.

        Returns
        -------
        tensap.FunctionalBasisArray
            The result of the matrix multiplication.

        '''
        data = np.matmul(self.data, v)
        return FunctionalBasisArray(data, self.basis, data.shape[1:])

    def matdiv(self, v):
        '''
        Compute the matrix multiplication of self.data with the inverse of v.

        Parameters
        ----------
        v : numpy.ndarray
            The array used in the matrix multiplication.

        Returns
        -------
        tensap.FunctionalBasisArray
            The result of the matrix multiplication.

        '''
        data = np.transpose(np.linalg.solve(np.transpose(v),
                                            np.transpose(self.data)))
        return FunctionalBasisArray(data, self.basis, self.shape)

    def dot(self, g, dim=None):
        '''
        Compute the dot product between the arrays self.data and g.data
        treated as collections of vectors. The function calculates
        the dot product of corresponding vectors along the first
        array dimension whose size does not equal 1.

        Parameters
        ----------
        g : tensap.FunctionalBasisArray
            The second object of the dot product.
        dim : int or list or numpy.ndarray, optional
            The dimension along which the dot product is computed. The default
            is None, indicating all the dimensions.

        Returns
        -------
        float or numpy.ndarray
            The result of the dot product.

        '''
        return np.sum(self.data * g.data, dim, keepdims=True)

    def norm(self, p='fro'):
        '''
        Compute the p-norm of the array self.data.

        See also numpy.linalg.norm.

        Parameters
        ----------
        p : int or numpy.inf or -numpy.inf or string, optional
            The order of the norm. The default is 'fro'.

        Returns
        -------
        float
            The norm of self.data.

        '''
        return np.linalg.norm(np.reshape(self.data,
                                         [self.basis.cardinal(), -1],
                                         order='F'), p)

    @staticmethod
    def is_random():
        '''
        Determine if the object is random.

        Returns
        -------
        bool
            Boolean equal to True if the object is random.

        '''
        return True

    def mean(self, measure=None):
        '''
        Compute the expectation of the function, according to the measure
        associated with the tensap.ProbabilityMeasure measure if provided, or
        to the standard tensap.ProbabilityMeasure associated with each
        polynomial if not.

        Parameters
        ----------
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used for the computation of the
            expectation. The default is None, indicating to use the standard
            tensap.ProbabilityMeasure associated with each polynomial.

        Returns
        -------
        numpy.ndarray
            The expectation of the function.

        '''
        M = self.basis.mean(measure)
        M = np.tile(np.ravel(M), np.concatenate((self.shape, [1])))
        M = np.transpose(M, np.concatenate(([np.ndim(M)-1],
                                            range(np.ndim(M)-1))))
        return np.sum(self.data*M, 0, keepdims=True)

    def expectation(self, measure=None):
        '''
        Compute the expectation of the function, according to the measure
        associated with the tensap.ProbabilityMeasure measure if provided, or
        to the standard tensap.ProbabilityMeasure associated with each
        polynomial if not.

        Parameters
        ----------
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used for the computation of the
            expectation. The default is None, indicating to use the standard
            tensap.ProbabilityMeasure associated with each polynomial.

        Returns
        -------
        numpy.ndarray
            The expectation of the function.

        '''
        return self.mean(measure)

    def variance(self, measure=None):
        '''
        Compute the variance of the function, according to the measure
        associated with the tensap.ProbabilityMeasure measure if provided, or
        to the standard tensap.ProbabilityMeasure associated with each
        polynomial if not.

        Parameters
        ----------
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used for the computation of the variance.
            The default is None, indicating to use the standard
            tensap.ProbabilityMeasure associated with each polynomial.

        Returns
        -------
        numpy.ndarray
            The variance of the function.

        '''
        m = self.expectation(measure)
        return self.dot_product_expectation(self, None, measure) - m**2

    def std(self, *args):
        '''
        Compute the standard deviation of the function, according to the
        measure associated with the tensap.ProbabilityMeasure measure if
        provided, or to the standard tensap.ProbabilityMeasure associated with
        each polynomial if not.

        Parameters
        ----------
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used for the computation of the standard
            deviation. The default is None, indicating to use the standard
            tensap.ProbabilityMeasure associated with each polynomial.

        Returns
        -------
        numpy.ndarray
            The standard deviation of the function.

        '''
        return np.sqrt(self.variance(*args))

    def dot_product_expectation(self, g, dims=None, measure=None):
        '''
        Compute the expectation of self(X)g(X), where X is the probability
        measure associated with the underlying basis, or measure if provided.

        For vector-valued functions of X, dims specifies the dimensions of
        self and g corresponding to the RandomVector measure.

        Parameters
        ----------
        g : tensap.FunctionalBasisArray
            The second function of the product.
        dims : list or numpy.ndarray, optional
            The dimensions of self and g corresponding to the RandomVector
            measure. The default is None.
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used for the computation of the
            expectation. The default is None, indicating to use the probability
            measure associated with the underlying basis.

        Raises
        ------
        NotImplementedError
            If the method is not implemented.

        Returns
        -------
        float or numpy.ndarray
            The result of the dot product.

        '''
        if dims is None:
            dims = range(self.basis.cardinal())
        if not (self.basis == g.basis) or not self.basis.is_orthonormal:
            raise NotImplementedError('Method not implemented.')

        return self.dot(g, 0)

    def norm_expectation(self, measure=None):
        '''
        Compute the L^2 norm of self(measure). If measure is not provided, use
        the probability measure associated with the underlying basis of self.

        Parameters
        ----------
        measure : tap.ProbabilityMeasure, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        float
            The L2 norm of the function.

        '''
        return np.sqrt(self.dot_product_expectation(self, None, measure))

    def conditional_expectation(self, dims, *args):
        '''
        Compute the conditional expectation of self with respect to
        the random variables dims (a subset of range(d)). The expectation
        with respect to other variables (in the complementary set of
        dims) is taken with respect to the probability measure given by
        tensap.RandomVector XdimsC if provided, or with respect to the
        probability measure associated with the corresponding bases of
        the function.

        Parameters
        ----------
        dims : numpy.ndarray
            The dimensions in which the expectation is computed.
        *args : tuple
            Additional parameters. See also the method conditional_expectation
            of the underlying basis.

        Returns
        -------
        f : tensap.FunctionalBasisArray
            The conditional expectation of the function.

        '''
        h = self.basis.conditional_expectation(dims, *args)
        f = deepcopy(self)
        f.data = np.matmul(h.data, np.reshape(self.data,
                                              [self.basis.cardinal(),
                                               np.prod(self.shape)],
                                              order='F'))
        f.data = np.reshape(f.data, (f.data.shape[0], f.shape), order='F')
        f.basis = h.basis
        return f

    def variance_conditional_expectation(self, alpha):
        '''
        Compute the variance of the conditional expectation of the function in
        dimensions in alpha.

        Parameters
        ----------
        alpha : numpy.ndarray
            The dimensions in which the variance of the conditional expectation
            of the function if computed.

        Returns
        -------
        numpy.ndarray
            The variance of the conditional expectation of the function.

        '''
        alpha = np.atleast_2d(alpha)
        m = self.expectation()
        v = np.zeros((np.shape(alpha)[0], np.prod(self.shape)))
        for i in range(np.shape(alpha)[0]):
            u = alpha[i, :]
            if np.all([isinstance(x, bool) for x in u]):
                u = np.nonzero(u)[0]

            if np.size(u) == 0:
                v[i, :] = 0
            else:
                mi = self.conditional_expectation(u)
                vi = mi.dot_product_expectation(mi) - m**2
                v[i, :] = np.ravel(vi)

        return np.reshape(v, np.concatenate(([np.shape(alpha)[0]],
                                             self.shape)), order='F')

    def eval(self, x, *args):
        if np.ndim(x) == 1:
            x = np.expand_dims(x, 1)
        H = self.basis.eval(x, *args)
        return self.eval_with_bases_evals(H, *args)

    def eval_with_bases_evals(self, H):
        '''
        Compute the evaluations of the function using the evaluations of the
        basis H.

        Parameters
        ----------
        H : numpy.ndarray
            The evaluations of the basis.

        Returns
        -------
        numpy.ndarray
            The evaluations of the function.

        '''
        y = np.matmul(H, np.reshape(self.data, [self.basis.cardinal(),
                                                np.prod(self.shape)],
                                    order='F'))
        return np.reshape(y, np.concatenate(([H.shape[0]], self.shape)),
                          order='F')

    def eval_derivative(self, n, x):
        '''
        Compute the n-derivative of the function at points x in R^d, with n a
        multi-index of size d.

        Parameters
        ----------
        n : int or list or numpy.ndarray
            The derivation order in all the dimensions, or the derivation
            orders for each dimension.
        x : numpy.ndarray
            The points used for the evaluation of the derivative.

        Raises
        ------
        NotImplementedError
            If the method is not implemented for the basis.

        Returns
        -------
        numpy.ndarray
            The evaluation of the n-derivative of the function at the points x.

        '''
        try:
            H = self.basis.eval_derivative(n, x)
            return self.eval_with_bases_evals(H)
        except Exception:
            raise NotImplementedError('Method not implemented for the basis.')

    def derivative(self, n):
        '''
        Compute the n-derivative of the function.

        Parameters
        ----------
        n : int
            The derivation order.

        Returns
        -------
        df : tensap.FunctionalBasisArray()
            The n-derivative of the function.

        '''
        df = deepcopy(self)
        df.basis = self.basis.derivative(n)
        return df

    def random(self, n=1, measure=None):
        '''
        Compute evaluations of the function at an array of points of size n,
        drawn randomly according to the tensap.ProbabilityMeasure measure if
        provided, or to the standard tensap.ProbabilityMeasure associated with
        each polynomial if not.

        Parameters
        ----------
        n : int, optional
            The number of random evaluations. The default is 1.
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used to draw the points of evaluation. The
            default is None, indicating to use the standard
            tensap.ProbabilityMeasure associated with each polynomial.

        Returns
        -------
        numpy.ndarray
            The random evaluations of the function.
        numpy.ndarray
            The points used for the evaluations of the function.

        '''
        fx, x = self.basis.random(n, measure)
        y = np.matmul(fx, np.reshape(self.data, [self.basis.cardinal(),
                                                 np.prod(self.shape)],
                                     order='F'))
        return np.reshape(y, np.concatenate(([fx.shape[0]], self.shape)),
                          order='F'), x

    def get_random_vector(self):
        '''
        Return the random vector associated with the basis functions of the
        object.

        Returns
        -------
        tensap.RandomVector
            The random vector associated with the basis functions of the
            object.

        '''
        return self.basis.get_random_vector()

    def storage(self):
        '''
        The storage complexity of the object.

        Returns
        -------
        int
            The storage complexity of the object.

        '''
        return np.size(self.data)

    def sparse_storage(self):
        '''
        The storage complexity of the object, taking into account the sparsity.

        Returns
        -------
        int
            The storage complexity of the object, taking into account the
            sparsity.

        '''
        return np.count_nonzero(self.data)

    def projection(self, basis, indices=None):
        '''
        Projection of the object on a functional basis using multi-indices
        indices if provided, or the multi-indices associated with
        the functional basis if not.

        Parameters
        ----------
        basis : tensap.FunctionalBasis (tensap.FullTensorProductFunctionalBasis
            or tensap.SparseTensorProductFunctionalBasis)
            The basis used for the projection.
        indices : tensap.MultiIndices, optional
            The multi-indices used for the projection. The default is None,
            indicating to use basis.indices.

        Returns
        -------
        g : FunctionalBasisArray
            The obtained projection.

        '''
        if indices is None:
            if isinstance(basis, tensap.SparseTensorProductFunctionalBasis):
                indices = basis.indices
            else:
                raise ValueError('Must specify a MultiIndices.')

        if self.basis.ndim() == basis.ndim() and \
                self.basis.cardinal() <= basis.cardinal():
            d = np.zeros((basis.cardinal(), np.prod(self.shape)))
            _, ia, ib = self.basis.indices.intersect_indices(indices)
            d[ib, :] = self.data[ia, :]
            if np.ndim(self.data) != np.ndim(d):
                d = np.reshape(d, [basis.cardinal(), self.shape], order='F')

            if isinstance(basis, tensap.FullTensorProductFunctionalBasis):
                H = tensap.FullTensorProductIntegrationRule(basis.bases)
            elif isinstance(basis, tensap.SparseTensorProductFunctionalBasis):
                H = tensap.SparseTensorProductFunctionalBasis(basis.bases,
                                                              indices)
            g = FunctionalBasisArray(d, H, self.shape)
        else:
            raise NotImplementedError('Method not implemented.')
        return g

    def sub_functional_basis(self):
        '''
        Converts the FunctionalBasisArray into a tensap.SubFunctionalbasis.

        Returns
        -------
        tensap.SubFunctionalbasis
            The FunctionalBasisArray as a tensap.SubFunctionalbasis.

        '''
        return tensap.SubFunctionalBasis(self.basis, self.data)

    def get_coefficients(self):
        '''
        Return the coefficients of the object.

        Returns
        -------
        numpy.ndarray
            The coefficients of the object.

        '''
        return np.reshape(self.data, np.concatenate(([self.basis.cardinal()],
                                                     self.shape)),
                          order='F')



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.


'''
Module magic_indices.

'''

import numpy as np


def magic_indices(F, n=None, option='left_right'):
    '''
    Return the set of n magic indices (i_1, j_1), ..., (i_n, j_n) constructed
    by a greedy algorithm.

    If option is 'left_right' (default value):
        [ik,jk] = arg max_{i,j} | F_ij - I_{k-1}(F)_ij |
        where I_{k-1}(F) is the rank-(k-1) matrix which interpolates F on the
        cross corresponding to rows (i_1, ..., i_{k-1}) and
        columns (j_1,..., j_{k-1}).

    If option is 'left':
        jk is equal to k and
        :math:`ik = arg max_{i} max_{1 <= j <= M} | F_ij - I_[k-1](F)_ij |`.

    If option is 'right':
        ik is equal to k and
        :math:`jk = arg max_{j} max_{1 <= i <= N} | F_ij - I[k-1](F)_ij |`

    Parameters
    ----------
    F : list or numpy.ndarray
        Input matrix.
    n : int, optional
        Number of magic indices. The default is None, indicating to choose
        n = np.min(F.shape).
    option : str, optional
        The selected option. The default is 'left_right'.

    Raises
    ------
    ValueError
        If the selected option is incorrect.

    Returns
    -------
    I, J : numpy.ndarray
        The magic indices.

    '''
    F = np.atleast_2d(F)
    if n is None:
        n = np.min(F.shape)

    if option == 'left_right':
        G = np.zeros(F.shape)
        I = []
        J = []
        for _ in range(n):
            i = np.argmax(np.max(np.abs(F - G), 1), 0)
            j = np.argmax(np.abs(F[i, :] - G[i, :]))
            I.append(i)
            J.append(j)
            G = np.matmul(F[:, J], np.linalg.solve(F[np.ix_(I, J)], F[I, :]))
    elif option == 'left':
        G = np.zeros(F.shape)
        I = []
        J = []
        for k in range(n):
            i = np.argmax(np.max(np.abs(F[:, :k+1] - G[:, :k+1]), 1), 0)
            I.append(i)
            J.append(k)
            G = np.matmul(F[:, J], np.linalg.solve(F[np.ix_(I, J)], F[I, :]))
    elif option == 'right':
        J, I = magic_indices(np.transpose(F), n, 'left')
    else:
        raise ValueError('Bad option.')
    return I, J



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module truncator.

'''

import warnings
import numpy as np
import tensap


class Truncator:
    '''
    Class Truncator.

    Attributes
    ----------
    tolerance : float
        The relative tolerance for the truncation.
    max_rank : int
        The maximum rank for the truncation.
    thresholding_type : str
        The thresholding type ('hard' or 'soft').
    thresholding_parameter : float
        The thresholding parameter.

    '''

    def __init__(self, tolerance=1e-8, max_rank=np.inf,
                 thresholding_type='hard', thresholding_parameter=None):
        '''
        Constructor for the class Truncator.

        Parameters
        ----------
        tolerance : float, optional
            The relative tolerance for the truncation. The default is 1e-8.
        max_rank : int, optional
            The maximum rank for the truncation. The default is numpy.inf.
        thresholding_type : str, optional
            The thresholding type ('hard' or 'soft'). The default is 'hard'.
        thresholding_parameter : float, optional
            The thresholding parameter. The default is None.

        Returns
        -------
        None.

        '''
        self.tolerance = tolerance
        self.max_rank = max_rank
        self.thresholding_type = thresholding_type
        self.thresholding_parameter = thresholding_parameter
        self._hsvd_type = 2  # 1 for root to leaves, 2 for leaves to root

    def truncate(self, tensor):
        '''
        Compute the truncation of the tensor with relative precision
        self.tolerance and maximal rank self.max_rank.

        Parameters
        ----------
        tensor : numpy.ndarray or tensap.FullTensor or tensorflow.Tensor or
        tensap.TreeBasedTensor
            The tensor to truncate.

        Raises
        ------
        NotImplementedError
            If the decomposition is not implemented for the tensor format.
        ValueError
            If the tensor is of order 1.

        Returns
        -------
        out : tensap.CanonicalTensor or tensap.TreeBasedTensor
            The truncated tensor.

        '''

        if not hasattr(tensor, 'order'):
            tensor = tensap.FullTensor(tensor)

        if tensor.order == 2:
            out = self.svd(tensor)
        elif tensor.order > 2:
            if isinstance(tensor, tensap.FullTensor):
                out = self.hosvd(tensor)
            elif isinstance(tensor, tensap.TreeBasedTensor):
                out = self.hsvd(tensor)
            else:
                raise NotImplementedError(
                    'Not implemented with this tensor format.')
        else:
            raise ValueError('Wrong tensor order.')
        return out

    def trunc_svd(self, matrix, tolerance=None, power=2):
        '''
        Compute the truncated svd of the matrix x with relative precision
        self.tolerance (or tolerance if provided) in Schatten p-norm (with p
        given by the input power) and maximal rank self.max_rank.

        Parameters
        ----------
        matrix : numpy.ndarray
            The matrix to truncate.
        tolerance : float, optional
            The relative tolerance for the truncation. The default is
            self.tolerance.
        power : int or 'inf' or float('inf') or numpy.inf, optional
            The integer p of the Schatten-p norm (1 <= p <= inf, p = 2 for
            Frobenius). The default is 2.

        Returns
        -------
        tensap.CanonicalTensor
            The truncated matrix.

        '''
        if tolerance is None:
            tolerance = self.tolerance

        left, sin_val, right = np.linalg.svd(matrix, full_matrices=False)
        if power in ('inf', float('inf'), np.inf):
            error = sin_val / np.max(sin_val)
        else:
            error = np.power(np.flip(np.cumsum(np.flip(np.power(sin_val,
                                                                power)))) /
                             np.sum(np.power(sin_val, power)), 1/power)
        error = np.concatenate((np.atleast_1d(error[1:]), [0]))
        ind = np.nonzero(error < tolerance)[0]
        if ind.size == 0:
            ind = np.min(matrix.shape) + 1
        else:
            ind = np.min(ind) + 1

        ind = np.int(np.min([ind, self.max_rank]))
        if self.thresholding_parameter is not None and \
                self.thresholding_parameter != 0:
            if self.thresholding_type == 'soft':
                sin_val -= self.thresholding_parameter
                ind = np.int(np.min([ind, np.nonzero(sin_val >= 0)[0][-1]+1]))
            elif self.thresholding_type == 'hard':
                ind = np.int(np.min([ind, np.nonzero(
                    sin_val >= self.thresholding_parameter)[0][-1]+1]))
        left = np.atleast_2d(left[:, :ind])
        sin_val = np.atleast_1d(sin_val[:ind])
        right = np.transpose(np.atleast_2d(right[:ind, :]))

        return tensap.CanonicalTensor([left, right], sin_val)

    def svd(self, tensor):
        '''
        Compute the truncated svd of an order-2 tensor.

        Parameters
        ----------
        tensor : numpy.ndarray or tensap.FullTensor or tensorflow.Tensor or
        tensap.TreeBasedTensor
            The tensor to truncate.

        Raises
        ------
        NotImplementedError
            If the method is not implemented.

        Returns
        -------
        out : tensap.CanonicalTensor or tensap.TreeBasedTensor
            The truncated tensor.

        '''
        assert tensor.ndim == 2, 'Wrong order.'

        if isinstance(tensor, np.ndarray):
            out = self.trunc_svd(tensor)
        elif isinstance(tensor, tensap.FullTensor):
            out = self.trunc_svd(tensor.data)
        elif isinstance(tensor, tensap.TreeBasedTensor):
            out = self.hsvd(tensor)
        elif isinstance(tensor, tensap.CanonicalTensor):
            tensor = tensor.orth()
            out = self.trunc_svd(tensor.core.data)
            out.space[0] = np.matmul(tensor.space[0], out.space[0])
            out.space[1] = np.matmul(tensor.space[1], out.space[1])
            out.shape = tensor.shape
        else:
            raise NotImplementedError('Method not implemented.')
        return out

    def hosvd(self, tensor):
        '''
        Compute the truncated hosvd of tensor.

        Parameters
        ----------
        tensor : np.ndarray or tensap.FullTensor or tensap.TreeBasedTensor
            The tensor to truncate.

        Raises
        ------
        ValueError
            If the input tensor is of the wrong type.

        Returns
        -------
        out : tensap.CanonicalTensor or tensap.TreeBasedTensor
            The truncated tensor.

        '''
        if isinstance(tensor, np.ndarray):
            tensor = tensap.FullTensor(tensor)

        order = tensor.order
        if order == 2:
            out = self.svd(tensor)
        else:
            max_rank = np.atleast_1d(self.max_rank)
            if max_rank.size == 1:
                max_rank = np.repeat(max_rank, order)
            local_tol = self.tolerance / np.sqrt(order)
            if isinstance(tensor, tensap.FullTensor):
                vec = np.empty(order, dtype=object)
                for dim in range(order):
                    self.max_rank = max_rank[dim]
                    vec[dim] = self.trunc_svd(tensor.matricize(dim).numpy(),
                                              local_tol)
                    vec[dim] = vec[dim].space[0]
                core = tensor.tensor_matrix_product(np.transpose(vec[0]), 0)
                for dim in np.arange(1, order):
                    core = core.tensor_matrix_product(
                        np.transpose(vec[dim]), dim)
                tensors = [core] + [tensap.FullTensor(x) for x in vec]
                tree = tensap.DimensionTree.trivial(order)
                out = tensap.TreeBasedTensor(tensors, tree)
            else:
                raise ValueError('Wrong type.')
        return out

    def hsvd(self, tensor, tree=None, is_active_node=None):
        '''
        Compute the truncated svd in tree-based tensor format of tensor.

        Parameters
        ----------
        tensor : tensap.FullTensor or tensap.TreeBasedTensor
            The tensor to truncate.
        tree : tensap.DimensionTree, optional
            The tree of the output tree-based tensor. The default is None,
            indicating if tensor is a tensap.TreeBasedTensor to take
            tensor.tree.
        is_active_node : numpy.ndarray, optional
            Logical array indicating if the nodes are active.. The default is
            None, indicating if tensor is a tensap.TreeBasedTensor to take
            tensor.is_active_node.

        Raises
        ------
        ValueError
            If the wrong value of the atttribude _hsvd_type is provided.
        NotImplementedError
            If the method is not implemented for the format.

        Returns
        -------
        out : tensap.TreeBasedTensor
            The truncated tensor in tree-based tensor format.

        '''
        if isinstance(tensor, tensap.TreeBasedTensor):
            if tree is not None or is_active_node is not None:
                warnings.warn('The provided tree and/or is_active_node '
                              'are not taken into account when x is a '
                              'tensap.TreeBasedTensor.')
            is_active_node = tensor.is_active_node
            tree = tensor.tree
        elif is_active_node is None:
            is_active_node = np.full(tree.nb_nodes, True)

        max_rank = np.atleast_1d(self.max_rank)
        if max_rank.size == 1:
            max_rank = np.repeat(max_rank, tree.nb_nodes)
            max_rank[tree.root-1] = 1

        local_tol = self.tolerance / np.sqrt(
            np.count_nonzero(is_active_node)-1)

        if isinstance(tensor, tensap.FullTensor):
            root_rank_greater_than_one = tensor.order == len(tree.dim2ind)+1

            tensors = np.empty(tree.nb_nodes, dtype=object)
            shape = np.array(tensor.shape)
            nodes_x = tree.dim2ind
            ranks = np.ones(tree.nb_nodes, dtype=int)

            for level in np.arange(np.max(tree.level), 0, -1):
                for nod in tree.nodes_with_level(level):
                    if is_active_node[nod-1]:
                        if tree.is_leaf[nod-1]:
                            rep = np.nonzero(nod == nodes_x)[0][0]
                        else:
                            children = tree.children(nod)
                            rep = [np.nonzero(np.isin(nodes_x, x))[0][0]
                                   for x in children]
                        rep_c = tensap.fast_setdiff(np.arange(nodes_x.size),
                                                    rep)

                        if root_rank_greater_than_one:
                            rep_c = np.concatenate((rep_c, [tensor.order-1]))

                        self.max_rank = max_rank[nod-1]
                        tmp = self.trunc_svd(tensor.matricize(rep).numpy(),
                                             local_tol)
                        tensors[nod-1] = tmp.space[0]
                        ranks[nod-1] = tensors[nod-1].shape[1]
                        shape_loc = np.hstack((shape[rep], ranks[nod-1]))
                        tensors[nod-1] = tensap.FullTensor(tensors[nod-1],
                                                           shape=shape_loc)
                        tmp = np.matmul(tmp.space[1],
                                        np.diag(tmp.core.data))
                        shape = np.hstack((shape[rep_c], ranks[nod-1]))
                        tensor = tensap.FullTensor(tmp, shape=shape)

                        if root_rank_greater_than_one:
                            perm = np.concatenate((np.arange(tensor.order-2),
                                                   [tensor.order-1],
                                                   [tensor.order-2]))
                            tensor = tensor.transpose(perm)
                            shape = shape[perm]
                            rep_c = rep_c[:-1]

                        nodes_x = np.hstack((nodes_x[rep_c], nod))
                    else:
                        tensors[nod-1] = []

            root_ch = tree.children(tree.root)
            rep = [np.nonzero(np.isin(nodes_x, x))[0][0] for x in root_ch]
            if root_rank_greater_than_one:
                rep = np.concatenate((rep, [tensor.order-1]))
            tensors[tree.root-1] = tensor.transpose(rep)
            out = tensap.TreeBasedTensor(tensors, tree)
        elif isinstance(tensor, tensap.TreeBasedTensor):
            if self._hsvd_type == 1:
                out = tensor.orth()
                gram = out.gramians()[0]
                mat = np.empty(gram.shape, dtype=object)
                shape = np.zeros(gram.shape)
                for nod in range(gram.size):
                    # Truncation of the Gramian in trace norm for a control
                    # of Frobenius norm of the tensor
                    if gram[nod] is not None:
                        self.max_rank = max_rank[nod]
                        tmp = self.trunc_svd(gram[nod], local_tol ** 2)
                        shape[nod] = tmp.core.shape[0]
                        mat[nod] = np.transpose(tmp.space[0])

                # Interior nodes without the root
                for level in np.arange(1, np.max(tree.level)):
                    nod_level = tensap.fast_setdiff(
                        tree.nodes_with_level(level),
                        np.nonzero(tree.is_leaf)[0]+1)
                    for nod in tree.nodes_indices[nod_level-1]:
                        order = out.tensors[nod-1].order
                        out.tensors[nod-1] = \
                            out.tensors[nod-1].tensor_matrix_product(
                                mat[nod-1], order-1)
                        parent = tree.parent(nod)
                        ch_nb = tree.child_number(nod)
                        out.tensors[parent-1] = \
                            out.tensors[parent-1].tensor_matrix_product(
                                mat[nod-1], ch_nb-1)

                # Leaves
                for nod in tree.dim2ind:
                    if out.is_active_node[nod-1]:
                        order = out.tensors[nod-1].order
                        out.tensors[nod-1] = \
                            out.tensors[nod-1].tensor_matrix_product(
                                mat[nod-1], order-1)
                        parent = tree.parent(nod)
                        ch_nb = tree.child_number(nod)
                        out.tensors[parent-1] = \
                            out.tensors[parent-1].tensor_matrix_product(
                                mat[nod-1], ch_nb-1)
                # Update the shape
                out = out.update_attributes()
                out.is_orth = False

            elif self._hsvd_type == 2:
                out = tensor.orth()
                gram = out.gramians()[0]
                for level in np.arange(np.max(tree.level), 0, -1):
                    for nod in tensap.fast_intersect(
                            tree.nodes_with_level(level), out.active_nodes):
                        # Truncation of the Gramian in trace norm for a control
                        # of Frobenius norm of the tensor
                        self.max_rank = max_rank[nod-1]
                        tmp = self.trunc_svd(gram[nod-1], local_tol ** 2)
                        tmp = np.transpose(tmp.space[0])
                        order = out.tensors[nod-1].order
                        out.tensors[nod-1] = \
                            out.tensors[nod-1].tensor_matrix_product(tmp,
                                                                     order-1)
                        parent = tree.parent(nod)
                        ch_nb = tree.child_number(nod)
                        out.tensors[parent-1] = out.tensors[parent-1].\
                            tensor_matrix_product(tmp, ch_nb-1)
                out = out.update_attributes()
                out.is_orth = True
                out.orth_node = tree.root
            else:
                raise ValueError('Wrong value of _hsvd_type.')
        else:
            raise NotImplementedError('Method not implemented.')
        return out

    def ttsvd(self, tensor):
        '''
        Compute the truncated svd in tensor-train format of tensor.

        Parameters
        ----------
        tensor : tensap.FullTensor or tensap.TreeBasedTensor
            The tensor to truncate.

        Returns
        -------
        tensap.TreeBasedTensor
            The truncated tensor in tree-based tensor format with a linear
            tree.

        '''
        tree = tensap.DimensionTree.linear(tensor.order)
        is_active_node = np.full(tree.nb_nodes, True)
        is_active_node[tree.dim2ind[1:]-1] = False

        return self.hsvd(tensor, tree, is_active_node)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on FunctionalBasis, projection, interpolation, least-squares
approximation.

'''

import numpy as np
import matplotlib.pyplot as plt
import tensap

# %% Interpolation on a polynomial space using Chebyshev Points
def FUN(x):
    # Function to approximate
    return np.cos(10*x)


FUN = tensap.UserDefinedFunction(FUN, 1)
FUN.evaluation_at_multiple_points = True
X = tensap.UniformRandomVariable(-1, 1)

# Interpolation basis and points
P = 30
H = tensap.PolynomialFunctionalBasis(X.orthonormal_polynomials(), range(P+1))
X_CP = tensap.chebyshev_points(H.cardinal(), [-1, 1])

# Interpolation of the function
F = H.interpolate(FUN, X_CP)

# Displays and error
print('Interpolation on a polynomial space using Chebyshev Points')
plt.figure()
X_PLOT = np.linspace(-1, 1, 100)
plt.plot(X_PLOT, FUN(X_PLOT))
plt.plot(X_PLOT, F(X_PLOT))
plt.legend(('True function', 'Interpolation'))
plt.show()
N = 100
ERR_L2, ERR_L_INF = F.test_error(FUN, N, X)
print('Mean squared error = %2.5e' % ERR_L2)

# %% Interpolation on a polynomial space using magic Points
def FUN(x):
    # Function to approximate
    return np.cos(10*x)


FUN = tensap.UserDefinedFunction(FUN, 1)
FUN.evaluation_at_multiple_points = True
X = tensap.UniformRandomVariable(-1, 1)

# Interpolation basis and points
P = 30
H = tensap.PolynomialFunctionalBasis(X.orthonormal_polynomials(), range(P+1))
X_MP = H.magic_points(X.random(10000))[0]

# Interpolation of the function
F = H.interpolate(FUN, X_MP)

# Displays and error
print('\nInterpolation on a polynomial space using magic points')
plt.figure()
X_PLOT = np.linspace(-1, 1, 100)
plt.plot(X_PLOT, FUN(X_PLOT))
plt.plot(X_PLOT, F(X_PLOT))
plt.legend(('True function', 'Interpolation'))
plt.show()
N = 100
ERR_L2, ERR_L_INF = F.test_error(FUN, N, X)
print('Mean squared error = %2.5e' % ERR_L2)

# %% Projection on polynomial space through quadrature
def FUN(x):
    # Function to approximate
    return x**2/2


FUN = tensap.UserDefinedFunction(FUN, 1)
FUN.evaluation_at_multiple_points = True
X = tensap.NormalRandomVariable()

# Integration rule
I = X.gauss_integration_rule(5)

# Approximation basis
P = 3
H = tensap.PolynomialFunctionalBasis(X.orthonormal_polynomials(), range(P+1))

# Computation of the projection
F = H.projection(FUN, I)

# Displays and error
print('\nProjection on polynomial space through quadrature')
N = 100
ERR_L2, ERR_L_INF = F.test_error(FUN, N, X)
print('Mean squared error = %2.5e' % ERR_L2)

# %% Interpolation on UserDefinedFunctionalBasis
def FUN(x):
    # Function to approximate
    return np.exp(-(x-1/2)**2)


FUN = tensap.UserDefinedFunction(FUN, 1)
FUN.evaluation_at_multiple_points = True
X = tensap.UniformRandomVariable(0, 1)

# Fourier Basis
N = 15
H = np.empty(2*N+1, dtype=object)
H[0] = lambda x: np.ones(np.size(x))
for i in np.arange(1, N+1):
    H[2*i-1] = lambda x, i=i: np.sqrt(2)*np.cos(2*np.pi*i*x)
    H[2*i] = lambda x, i=i: np.sqrt(2)*np.sin(2*np.pi*i*x)
H = tensap.UserDefinedFunctionalBasis(H, X)

# Display of the Fourier basis
plt.figure()
X_PLOT = np.linspace(0, 1, 100)
plt.plot(X_PLOT, H.eval(X_PLOT))
plt.title('Fourier basis')
plt.show()

# Computation of the interpolation
X_MP = H.magic_points(X.random(10000))[0]
F = H.interpolate(FUN, X_MP)

# Displays and error
print('\nInterpolation on UserDefinedFunctionalBasis')
plt.figure()
X_PLOT = np.linspace(0, 1, 100)
plt.plot(X_PLOT, FUN(X_PLOT))
plt.plot(X_PLOT, F(X_PLOT))
plt.legend(('True function', 'Interpolation'))
plt.show()
N = 100
ERR_L2, ERR_L_INF = F.test_error(FUN, N, X)
print('Mean squared error = %2.5e' % ERR_L2)

# %% Interpolation with a radial basis
def FUN(x):
    # Function to approximate
    return np.exp(-x**2)


FUN = tensap.UserDefinedFunction(FUN, 1)
FUN.evaluation_at_multiple_points = True
X = tensap.UniformRandomVariable(-1, 1)

# Radial basis
N = 20
X_LIN = np.linspace(-1, 1, N)
S = 10/N
k = lambda x, y, s=S: np.exp(-(x-y)**2/s**2)
H = np.empty(N, dtype=object)
for i in range(N):
    H[i] = lambda y, x=X_LIN[i], k=k: k(y, x)
H = tensap.UserDefinedFunctionalBasis(H, X)

# Computation of the interpolation
X_MP = H.magic_points(X.random(100000))[0]
F = H.interpolate(FUN, X_MP)

# Displays and error
print('\nInterpolation with a radial basis')
plt.figure()
X_PLOT = np.linspace(-1, 1, 100)
plt.plot(X_PLOT, FUN(X_PLOT))
plt.plot(X_PLOT, F(X_PLOT))
plt.legend(('True function', 'Interpolation'))
plt.show()
N = 100
ERR_L2, ERR_L_INF = F.test_error(FUN, N, X)
print('Mean squared error = %2.5e' % ERR_L2)

# %% Projection on polynomial space through quadrature
def FUN(x):
    # Function to approximate
    return x**2/2


FUN = tensap.UserDefinedFunction(FUN, 1)
FUN.evaluation_at_multiple_points = True
X = tensap.NormalRandomVariable()

# Integration rule
I = X.gauss_integration_rule(5)

# Approximation basis
P = 3
H = tensap.PolynomialFunctionalBasis(tensap.HermitePolynomials(), range(P+1))

# Computation of the approximation
F = H.projection(FUN, I)

# Derivative and second derivative of F through projection
DF = H.projection(lambda x, F=F: F.eval_derivative(1, x), I)
DDF = H.projection(lambda x, DF=DF: DF.eval_derivative(1, x), I)

# Displays and error
print('\nProjection on polynomial space through quadrature')
N = 100
ERR_L2, ERR_L_INF = F.test_error(FUN, N, X)
print('Mean squared error = %2.5e' % ERR_L2)
plt.figure()
X_PLOT = np.linspace(-1, 1, 100)
plt.plot(X_PLOT, F(X_PLOT))
plt.plot(X_PLOT, DF(X_PLOT))
plt.plot(X_PLOT, DDF(X_PLOT))
plt.legend(('F', 'df', 'ddf'))
plt.show()

# %% Least-squares approximation
def FUN(x):
    # Function to approximate
    return np.exp(-(x-1/2)**2)


FUN = tensap.UserDefinedFunction(FUN, 1)
FUN.evaluation_at_multiple_points = True
X = tensap.UniformRandomVariable(0, 1)

# Approximation basis: Hermite polynomials of maximal degree P
P = 10
H = tensap.PolynomialFunctionalBasis(tensap.HermitePolynomials(), range(P+1))

# Solver
SOLVER = tensap.LinearModelLearningSquareLoss()
SOLVER.regularization = False
SOLVER.basis_adaptation = False

# Training sample
X_TRAIN = X.random(100)
Y_TRAIN = FUN(X_TRAIN)

# Computation of the approximation
SOLVER.basis = H
SOLVER.training_data = [X_TRAIN, Y_TRAIN]
F, OUTPUT = SOLVER.solve()

# Displays and error
print('\nLeast-squares approximation')
plt.figure()
X_PLOT = np.linspace(-1, 1, 100)
plt.plot(X_PLOT, FUN(X_PLOT))
plt.plot(X_PLOT, F(X_PLOT))
plt.legend(('True function', 'Interpolation'))
plt.show()
X_TEST = X.random(100)
F_X_TEST = F(X_TEST)
Y_TEST = FUN(X_TEST)
print('Mean squared error = %2.5e' %
      (np.linalg.norm(Y_TEST-F_X_TEST) / np.linalg.norm(Y_TEST)))



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module dimension_tree.

'''

import numpy as np


class DimensionTree:
    '''
    Class DimensionTree.

    Attributes
    ----------
    adjacency_matrix : numpy.ndarray
        Adjacency matrix. The nth row indicates the sons of the (n+1)th node.
        The nth column indicates the parents of the (n+1)th node.
    arity : int
        Maximal number of children.
    _child_number : numpy.ndarray
        _child_number[n] = k means that (n+1)th node is the kth child of its
        parent.
    _children : numpy.ndarray
        _children[:,n] is the set of indices of the children of (n+1)th node.
    dim2ind : numpy.ndarray
        dim2ind[k] is the index of the node (leaf) corresponding to
        dimension k.
    dims : list
        dims[n] is the set of dimensions associated with node n+1.
    internal_nodes : numpy.ndarray
        Indices of internal (non leaf) nodes.
    is_leaf : numpy.ndarray
        is_leaf[n] = True if the (n+1)th node is a leaf and False otherwise.
    level : numpy.ndarray
        level[n] is the level of the (n+1)th node.
    nb_nodes : int
        Number of nodes.
    nodes_parent_of_leaves : numpy.ndarray
        Indices of nodes which are parents of leaves.
    _parent : numpy.ndarray
        _parent[n] is the index of the parent of the (n+1)th node.
    root : numpy.int64
        Index of the root node.
    sibling : numpy.ndarray
        sibling[:,n] contains the indices of the children of the parent of
        the (n+1)th node.
    nodes_indices : numpy.ndarray
        Indices of the nodes of the tree.
    plot_options : dict
        Options for plotting the tree.

    '''

    def __init__(self, dim2ind, adjacency_matrix):
        '''
        Constructor of the class DimensionTree.

        Create a dimension partition tree over D = {1,...,d} from an
        adjacency matrix.

        Parameters
        ----------
        dim2ind : numpy.ndarray or list
            dim2ind[k] is the index of the node (leaf) corresponding to
            dimension k.
        adjacency_matrix : numpy.ndarray or list of a list
            The nth row indicates the sons of the (n+1)th node. The nth column
            indicates the parents of the (n+1)th node.

        Returns
        -------
        None.

        '''
        self.dim2ind = np.array(dim2ind)
        self.adjacency_matrix = np.array(adjacency_matrix)
        self.plot_options = {'level_alignment': False}

        self._precompute_attributes()
        self.update_dims_from_leaves()

    def __repr__(self):
        return ('<DimensionTree:{n}' +
                '{t}nb_nodes = {},{n}' +
                '{t}arity = {},{n}' +
                '{t}dim2ind = {},{n}' +
                '{t}is_leaf = {},{n}').format(self.nb_nodes,
                                              self.arity,
                                              self.dim2ind,
                                              self.is_leaf,
                                              t='\t', n='\n')

    def permute(self, sigma):
        '''
        Permute the dimensions of the DimensionTree.

        Parameters
        ----------
        sigma : list
            Indicates the permutation of the dimensions.

        Returns
        -------
        DimensionTree
            A DimensionTree with permuted dimensions.

        '''
        dim2ind = np.copy(self.dim2ind)
        dim2ind[sigma] = dim2ind.flatten()
        return DimensionTree(dim2ind, np.copy(self.adjacency_matrix))

    def ipermute(self, sigma):
        '''
        Inverse permutation of the dimensions of the DimensionTree.

        Parameters
        ----------
        sigma : list
            Indicates the permutation of the dimensions.

        Returns
        -------
        DimensionTree
            A DimensionTree with permuted dimensions.

        '''
        dim2ind = np.copy(self.dim2ind)
        dim2ind = dim2ind[sigma].flatten()
        return DimensionTree(dim2ind, np.copy(self.adjacency_matrix))

    def __eq__(self, T):
        is_equal = (len(self.dim2ind) == len(T.dim2ind)) and \
            (self.nb_nodes == T.nb_nodes) and \
            (np.max(self.level) == np.max(T.level))

        if is_equal:
            for level in np.arange(1, np.max(self.level)+1):
                nodes1 = self.nodes_with_level(level)
                nodes2 = T.nodes_with_level(level)
                if len(nodes1) != len(nodes2):
                    return is_equal
                dims1 = [self.dims[i] for i in nodes1-1]
                dims2 = [T.dims[i] for i in nodes2-1]
                for i, dim1 in enumerate(dims1):
                    for dim2 in dims2:
                        if np.array_equal(np.sort(dim1),
                                          np.sort(dim2)):
                            nodes1[i] = 0
                            break
                if any(nodes1 != 0):
                    is_equal = False
                    return is_equal
        return is_equal

    def __ne__(self, T):
        return not self == T

    def children(self, nod):
        '''
        Return the children of a given node.

        Parameters
        ----------
        nod : int
            Node for which to compute the children.

        Returns
        -------
        list
            List of children of nod.

        '''
        children = self._children[:, nod-1]
        return children[children != 0]

    def child_number(self, nod):
        '''
        Return the child number of nod.

        Parameters
        ----------
        nod : int
            Node for which to compute the child number.

        Returns
        -------
        int
            The child number.

        '''
        return self._child_number[nod-1]

    def parent(self, nod):
        '''
        Return the parent of a given node.

        Parameters
        ----------
        nod : int
            Node for which to compute the parent.

        Returns
        -------
        int
            The parent of nod.

        '''
        return self._parent[nod-1]

    def ascendants(self, nod):
        '''
        Return the ascendants of a given node.

        Parameters
        ----------
        nod : int
            Node for which to compute the ascendants.

        Returns
        -------
        anod : list
            List of ascendants of nod.

        '''
        anod = []
        pnod = self.parent(nod)
        while pnod:
            anod.append(pnod)
            pnod = self.parent(pnod)
        return np.array(anod)

    def descendants(self, nod):
        '''
        Return the descendants of a given node.

        Parameters
        ----------
        nod : int
            Node for which to compute the descendants.

        Returns
        -------
        dnod : list
            List of descendants of nod.

        '''
        dnod = np.array([], dtype=int)
        chnod = self.children(nod)
        chnod = chnod[chnod != 0]
        if len(chnod) != 0:
            for children in chnod:
                dnod = np.concatenate((dnod, [children],
                                       self.descendants(children)))
        return np.array(dnod)

    def sub_dimension_tree(self, root):
        '''
        Extract a sub dimension tree.

        The attribute dim2ind of the sub dimension tree gives the nodes indices
        corresponding to the dimensions in T.dims[r] (not sorted).

        Parameters
        ----------
        root : int
            Index of the node which is the root of the sub dimension tree.

        Returns
        -------
        DimensionTree
            Sub dimension tree.
        nod : np.array
            Extracted nodes from T.

        '''
        dims = self.dims[root-1]
        nod = np.concatenate(([root], self.descendants(root)))
        adj_mat = np.copy(self.adjacency_matrix[nod[:, None]-1, nod-1])
        dim2ind = np.nonzero(np.isin(nod, self.dim2ind[dims]))[0] + 1
        return DimensionTree(dim2ind, adj_mat), nod

    def node_with_dims(self, dims):
        '''
        Return the index of the node with given set of dimensions.

        Return the index of the node corresponding to dimensions dims or an
        empty array if no node contains these dimensions.

        Parameters
        ----------
        dims : list or numpy.ndarray
            List of dimensions.

        Returns
        -------
        numpy.ndarray
            Index of the node with the given set of dimensions dims.

        '''
        dims = np.sort(dims)
        nod = [np.array_equal(dims, np.sort(x)) for x in self.dims]
        return np.nonzero(nod)[0]+1

    def update_dims_from_leaves(self):
        '''
        Update the dimensions of all nodes from the dimensions of the leaves
        given in T.dim2ind.

        Returns
        -------
        DimensionTree
            The DimensionTree object with updated attribute dims.

        '''
        _dims = np.zeros(self.nb_nodes, dtype=int)
        _dims[self.dim2ind-1] = np.arange(len(self.dim2ind))
        _dims = np.split(_dims, len(_dims))

        for level in np.arange(np.max(self.level), -1, -1):
            nod_level = self.nodes_with_level(level)
            for nod in nod_level:
                if not self.is_leaf[nod-1]:
                    children = self._children[:, nod-1]
                    _dims[nod-1] = np.hstack([_dims[i-1] for i in
                                              children[children != 0]])

        self.dims = _dims
        return self

    def nodes_with_level(self, level):
        '''
        Return the indices of the nodes at a given level.

        Parameters
        ----------
        level : int
            Level.

        Returns
        -------
        numpy.ndarray
            Nodes with given level.

        '''
        ind_lvl = np.repeat(False, self.nb_nodes)
        ind_lvl[self.level == level] = True
        return np.nonzero(ind_lvl)[0]+1

    def tree_layout(self):
        '''
        Return the layout of a tree.

        Returns
        -------
        dict
            The layout of the tree, used for plotting.

        '''
        def post_order(T, root=None):
            '''
            Return the post-ordering of a tree.

            Parameters
            ----------
            T : tensap.DimensionTree
            The dimension tree.
            root : int, optional
                The root of the current subtree. The default is None,
                indicating to use T.root.

            Returns
            -------
            post : numpy.ndarray
                The post-ordering of the tree or of one of its subtrees.

            '''
            if root is None:
                root = T.root
            post = np.array([], dtype=int)

            ch = T.children(root)
            if ch.size:
                for nod in ch:
                    post = np.concatenate((post, post_order(T, nod)))
            post = np.concatenate((post, ch))
            return post

        post = post_order(self)

        xmin = np.full(self.nb_nodes, self.nb_nodes)
        xmax = np.zeros(self.nb_nodes)
        y = np.zeros(self.nb_nodes)
        n_leaves = 0
        for nod in post:
            if self.is_leaf[nod-1]:
                n_leaves += 1
                xmin[nod-1] = n_leaves
                xmax[nod-1] = n_leaves
                y[nod-1] = 0
            pa = self.parent(nod)
            xmin[pa-1] = np.min((xmin[pa-1], xmin[nod-1]))
            xmax[pa-1] = np.max((xmax[pa-1], xmax[nod-1]))
            y[pa-1] = np.max((y[pa-1], y[nod-1]+1))

        x = (xmin + xmax) / 2

        x /= (self.dim2ind.size + 1)
        y = (y+1) / (np.max(y)+2)

        if self.plot_options['level_alignment']:
            height = np.max(y) - np.min(y)
            for alpha in np.arange(1, self.nb_nodes+1):
                y[alpha-1] = np.max(y) - self.level[alpha-1] * \
                    height / np.max(self.level)

        pos = [np.array([xx, yy]) for xx, yy in zip(x, y)]
        return dict(zip(range(self.nb_nodes), pos))

    def plot(self, **args):
        '''
        Plot the tree with the nodes indices.

        This method requires the package igraph.

        Parameters
        ----------
        node_color : str, optional
            Color for the colored nodes. The default is 'red'.
        colored_nodes : list or numpy.ndarray, optional
            Colored nodes. The default is [].

        Returns
        -------
        None.

        '''
        self.plot_with_labels_at_nodes(self.nodes_indices, **args)

    def plot_with_labels_at_nodes(self, labels, node_color='red',
                                  colored_nodes=None, title=None):
        '''
        Plot the tree with labels at nodes.

        This method requires the package igraph.

        Parameters
        ----------
        labels : list or numpy.ndarray
            Nodes labels.
        node_color : str, optional
            Color for the colored nodes. The default is 'red'.
        colored_nodes : list or numpy.ndarray, optional
            Colored nodes. The default is [].
        title : str, optional
            The title of the plot. The default is None.

        Returns
        -------
        None.

        '''
        import networkx as nx
        import matplotlib.pyplot as plt

        plt.figure()
        g = nx.convert_matrix.from_numpy_matrix(self.adjacency_matrix)
        pos = self.tree_layout()
        nx.draw_networkx_nodes(g, pos,
                               nodelist=np.setdiff1d(
                                   np.arange(1, self.nb_nodes+1),
                                   colored_nodes)-1,
                               node_color='w',
                               edgecolors='k',
                               node_size=100)

        if colored_nodes is not None:
            nx.draw_networkx_nodes(g, pos,
                                   nodelist=colored_nodes-1,
                                   node_color=node_color,
                                   edgecolors='k',
                                   node_size=100)

        nx.draw_networkx_edges(g, pos)
        labels = dict(zip(range(self.nb_nodes), [x if x is not None
                                                 else '' for x in labels]))
        y = [x[1] for x in pos.values()]
        min_height = np.min(y)
        shift = (np.max(y)-np.min(y))/np.max(self.level)/2
        pos_labels = {}
        for k, v in pos.items():
            if v[1] == min_height:
                pos_labels[k] = (v[0], v[1]-shift)
            else:
                pos_labels[k] = (v[0], v[1]+shift)
        nx.draw_networkx_labels(g, pos_labels, labels)

        plt.axis('off')
        plt.title(title)
        axes = plt.gca()
        axes.set_ylim([0, 1])
        plt.show()

    def plot_dims(self, nodes=None, **args):
        '''
        Plot the dimensions associated with the nodes of the tree.

        This method requires the package igraph.

        Parameters
        ----------
        nodes : list or numpy.ndarray, optional
            List of leaf nodes for which to display the dimensions. The
            default is None, the dimensions of all the leaf nodes are
            displayed.
        Returns
        -------
        None.

        '''
        if nodes is None:
            nodes = self.dim2ind
        labels = [np.nonzero(nodes == i+1)[0][0] if self.is_leaf[i] else None
                  for i in range(self.nb_nodes)]
        self.plot_with_labels_at_nodes(labels=labels, **args)

    def _precompute_attributes(self):
        '''
        Precompute the attributes of the DimensionTree from the attributes
        adjacency_matrix and dim2ind.

        Returns
        -------
        DimensionTree
            A DimensionTree with updated attributes.

        '''
        self.nb_nodes = self.adjacency_matrix.shape[0]
        self.nodes_indices = np.arange(1, self.nb_nodes+1)
        self.arity = np.max(np.sum(self.adjacency_matrix, 1))
        self._parent = np.matmul(range(1, self.nb_nodes+1),
                                 self.adjacency_matrix)
        self.root = self.nodes_indices[self._parent == 0][0]
        self.is_leaf = np.repeat(False, self.nb_nodes)
        self.is_leaf[self.dim2ind-1] = True

        _children = np.zeros([self.arity, self.nb_nodes], dtype=int)
        for i in range(self.nb_nodes):
            ind = np.nonzero(self.adjacency_matrix[i, :])[0]
            _children[0:len(ind), i] = ind+1
        self._children = _children

        _child_number = np.zeros(self.nb_nodes, dtype=int)
        for i in range(self.nb_nodes):
            if self._parent[i]:
                ind1 = self._children[:, self._parent[i]-1]
                ind2 = np.nonzero(ind1 == i+1)
                if ind2[0].size:
                    _child_number[i] = ind2[0][0]+1
        self._child_number = _child_number

        _sibling = np.zeros([self.arity, self.nb_nodes], dtype=int)
        for i in range(self.nb_nodes):
            if self._parent[i]:
                _sibling[:, i] = self._children[:, self._parent[i]-1]
        self.sibling = _sibling

        _level = np.zeros(self.nb_nodes, dtype=int)
        ind = self._children[:, self.root-1]
        ind = ind[ind != 0]
        level = 1
        while ind.size:
            _level[ind-1] = level
            ind = np.reshape(self._children[:, ind-1], [1, -1])
            ind = ind[np.nonzero(ind)]
            level += 1
        self.level = _level

        self.internal_nodes = np.setdiff1d(self.nodes_indices, self.dim2ind)
        self.nodes_parent_of_leaves = np.unique(self._parent[self.dim2ind-1])
        return self

    @staticmethod
    def trivial(order):
        '''
        Create a dimension tree with one level.

        Parameters
        ----------
        order : int
            Order of the tensor (dimension).

        Returns
        -------
        DimensionTree
            Trivial dimension tree.

        '''
        d2i = np.arange(2, order+2)
        adj_mat = np.zeros([order+1]*2, dtype=int)
        adj_mat[0, 1:] = 1
        return DimensionTree(d2i, adj_mat)

    @staticmethod
    def linear(order):
        '''
        Create a linear dimension tree.

        Parameters
        ----------
        order : int
            Order of the tensor (dimension).

        Returns
        -------
        DimensionTree
            Linear dimension tree.

        '''
        order = np.array(order)
        if order.size == 1:
            dim = order
            order = np.arange(dim)
        else:
            dim = order.size
            order -= 1

        d2i = np.concatenate(([2*dim-2], 2*np.arange(dim-1, 0, -1)+1))
        adj_mat = np.zeros([2*dim-1]*2, dtype=int)
        adj_mat[0, 1:3] = 1
        for level in range(dim-2):
            adj_mat[2*level+1, 2*level+3] = 1
            adj_mat[2*level+1, 2*level+4] = 1
        d2i[order] = d2i.flatten()
        return DimensionTree(d2i, adj_mat)

    @staticmethod
    def balanced(order):
        '''
        Create a balanced dimension tree.

        Parameters
        ----------
        order : int
            Order of the tensor (dimension).

        Returns
        -------
        DimensionTree
            Balanced dimension tree.

        '''
        order = np.array(order)
        if order.size == 1:
            dim = order
            order = range(dim)
        else:
            dim = order.size
            order -= 1

        d2i = np.arange(dim, 2*dim)
        adj_mat = np.zeros([2*dim-1]*2, dtype=int)
        adj_mat[0, 1:3] = 1
        for i in range(1, dim-1):
            adj_mat[i, 2*i+1:2*i+3] = 1
        d2i[order] = d2i.flatten()
        return DimensionTree(d2i, adj_mat)

    @staticmethod
    def random(order, arity=2):
        '''
        Create a random dimension tree over {1,...,order}.

        If arity is an interval [amin,amin], then the number of children of a
        node is randomly drawn from the uniform distribution over
        {amin,...,amax}.

        Parameters
        ----------
        order : int
            Order of the tensor (dimension).
        arity : int, list or numpy.ndarray, optional
            Arity or interval for the arity. The default is 2.

        Returns
        -------
        DimensionTree
            Random dimension tree.

        '''
        arity = np.array(arity)
        if arity.size == 1:
            arity = np.repeat(arity, 2)
        arity[1] += 1
        dims_nodes = [np.arange(order)]
        nb_nodes = 1
        new_nodes = [1]
        adj_mat = np.zeros([2*order]*2, dtype=int)
        dim2ind = np.zeros(order, dtype=int)

        while len(new_nodes) != 0:
            parent_nodes = new_nodes
            new_nodes = []
            for pnod in parent_nodes:
                dims = dims_nodes[pnod-1]
                if len(dims) == 1:
                    dim2ind[dims-1] = pnod
                else:
                    pnod_arity = np.min(np.append(np.random.randint(*arity),
                                                  dims.size))
                    for k in range(pnod_arity):
                        nb_nodes += 1
                        new_nodes.append(nb_nodes)
                        adj_mat[pnod-1, nb_nodes-1] = 1
                        if k == pnod_arity-1:
                            dims_nodes.append(dims)
                            dims = []
                        else:
                            n_children = np.max(np.append(np.random.randint(
                                len(dims)-pnod_arity+k+2), 1))
                            perm = np.random.\
                                permutation(len(dims))[:n_children]
                            dims_nodes.append(dims[perm])
                            dims = np.delete(dims, perm)
                        if len(dims) == 0:
                            break
        adj_mat = adj_mat[:nb_nodes, :nb_nodes]
        return DimensionTree(dim2ind, adj_mat)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module utils.

'''

import numpy as np


def fast_intersect(a, b):
    aux = np.concatenate((a, b))
    aux.sort()
    mask = aux[1:] == aux[:-1]
    return aux[:-1][mask]


def fast_setdiff(a, b):
    return a[np.logical_not(np.isin(a, b, assume_unique=True))]


def integer2baseb(i, b, d=None):
    '''
    Returns the representation [i_1, ..., i_d] in base b of a set of non
    negative integers i = sum_{k=1}^d i_k b^(d-k) in [0, b^d-1].

    Parameters
    ----------
    i : list or numpy.ndarray
        The integers to be convertes in base b.
    b : int
        The base.
    d : int, optional
        The dimension. The default is the minimal integer allowing the
        representation of max(i).

    Returns
    -------
    numpy.ndarray
        The representation [i_1, ..., i_d] in base b of the integers.

    '''
    if d is None:
        d = int(np.ceil(np.log(np.max(i)+1)/np.log(b)))

    I = np.unravel_index(np.ravel(i), np.full(d, b), order='F')
    return np.fliplr(np.transpose(np.array(I)))


def baseb2integer(I, b):
    '''
    Return the integers with given representations in base b.

    Parameters
    ----------
    I : numpy.ndarray or list
        A row of I contains d integers [i_1, ..., i_d] in {0, ..., b-1}
        associated with an integer i = sum_{k=1}^d i_k b^(d-k) in [0, b^d-1].
    b : int
        The base.

    Returns
    -------
    numpy.ndarray
        The integers with given representations in base b.

    '''
    I = np.atleast_2d(I)
    d = I.shape[1]
    I = np.fliplr(I)
    I = np.transpose(I)
    return np.ravel_multi_index(I, np.full(d, b), order='F')



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module learning.

'''

import tensap


class Learning:

    def __init__(self, loss):
        assert isinstance(loss, tensap.LossFunction), \
            'Must provide a tensap.LossFunction object.'

        self.loss_function = loss
        self.display = True
        self.model_selection = True
        self.model_selection_options = {'stop_if_error_increase': False}
        self.error_estimation = False
        self.error_estimation_type = 'leave_out'
        self.error_estimation_options = {'correction': True}
        self.training_data = None
        self.test_error = False
        self.test_data = None

    @staticmethod
    def linear_model(loss_function):
        if isinstance(loss_function, tensap.SquareLossFunction):
            model = tensap.LinearModelLearningSquareLoss()
        elif isinstance(loss_function, tensap.DensityL2LossFunction):
            model = tensap.LinearModelLearningDensityL2()
        elif isinstance(loss_function, tensap.CustomLossFunction):
            model = tensap.LinearModelLearningCustomLoss(loss_function)
        return model

    def risk_estimation(self, fun, sample, *args):
        return self.loss_function.risk_estimation(fun, sample, *args)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module random_vector.

'''

from copy import deepcopy
import numpy as np
import tensap


class RandomVector(tensap.ProbabilityMeasure):
    '''
    Class RandomVector.

    Attributes
    ----------
    random_variables : numpy.ndarray
        The RandomVariable objects constituting the random vector.
    copula : tensap.Copula
        The copula of the random vector.

    '''

    def __init__(self, random_variables, order=None,
                 copula=tensap.IndependentCopula()):
        if isinstance(random_variables, tensap.RandomVector):
            random_variables = random_variables.random_variables

        if order is not None and isinstance(random_variables,
                                            tensap.RandomVariable):
            random_variables = [random_variables] * order

        self.random_variables = np.atleast_1d(deepcopy(random_variables))
        self.copula = copula

    def __repr__(self):
        return ('<{}:{n}' +
                '{t}random_variables = {},{n}' +
                '{t}copula = {}>').format(self.__class__.__name__,
                                          self.random_variables,
                                          self.copula.__class__.__name__,
                                          t='\t', n='\n')

    def marginal(self, ind):
        if not isinstance(self.copula, tensap.IndependentCopula):
            raise NotImplementedError(
                'Not implemented for non IndenpendentCopula.')
        return RandomVector(self.random_variables[ind])

    def ndim(self):
        return np.sum([x.ndim() for x in self.random_variables])

    @property
    def size(self):
        '''
        Return the number of random variables constituting the RandomVector.

        Returns
        -------
        int
            The number of random variables constituting the RandomVector.

        '''
        return len(self.random_variables)

    def __eq__(self, rv_2):
        if not (isinstance(self, tensap.RandomVector) and
                isinstance(rv_2, tensap.RandomVector)):
            is_equal = False
        elif len(self.random_variables) != len(rv_2.random_variables):
            is_equal = False
        else:
            is_equal = True
            for ind in zip(self.random_variables, rv_2.random_variables):
                is_equal = is_equal and (ind[0] == ind[1])
        return is_equal

    def __neq__(self, rv_2):
        return not (self == rv_2)

    def support(self):
        assert isinstance(self.copula, tensap.IndependentCopula), \
            'Not implemented for non IndenpendentCopula.'
        return [x.support() for x in self.random_variables]

    def truncated_support(self):
        '''
        Return the truncated support of the random vector.

        Raises
        ------
        NotImplementedError
            If the copula is not an IndependentCopula.

        Returns
        -------
        sup : numpy.ndarray
            The truncated support of the random vector.

        '''
        assert isinstance(self.copula, tensap.IndependentCopula), \
            'Not implemented for non IndependentCopula.'
        return [x.truncated_support() for x in self.random_variables]

    def get_standard_random_vector(self):
        '''
        Return the standard RandomVector associated with self.

        Returns
        -------
        tensap.RandomVector
            The standard RandomVector.

        '''
        return RandomVector([x.get_standard_random_variable() for x in
                             self.random_variables])

    def iso_probabilistic_grid(self, n):
        '''
        Generate a grid of (n[0]-1) x ... x (n[d-1]-1) points
        (x_{i_1}^1, ..., x_{i_d}^d) such that the N = (n[0] x ... x(n[d-1]
        sets [x_{i_1-1}^1, x_{i_1}^1] x ... x [x_{i_d-1}^1, x_{i_d}^1] have
        the same probability p = 1/N.

        Parameters
        ----------
        n : int
            The number of points of the grid plus one in each or all the
            dimensions.

        Returns
        -------
        tensap.FullTensorGrid
            The iso-probabilistic grid.

        '''
        assert isinstance(self.copula, tensap.IndependentCopula), \
            'The method only works for independent copulas.'

        d = self.size
        if np.size(n) == 1 and n < 1:
            n = int(np.ceil(n**(-1/d)))

        if np.size(n) != d:
            n = np.full(d, n)

        b = [x.iso_probabilistic_grid(y) for x, y in zip(self.random_variables,
                                                         n)]
        return tensap.FullTensorGrid(b)

    def lhs_random(self, n):
        '''
        Latin Hypercube Sampling of the RandomVector of n points.

        Requires the package pyDOE.

        Parameters
        ----------
        n : int
            Number of points.

        Returns
        -------
        list
            List containing the coordinates of the Latin Hypercube Sampling in
            each dimension.

        '''
        from pyDOE import lhs
        A = lhs(self.size, samples=n)
        A = RandomVector(tensap.UniformRandomVariable(0, 1),
                         self.size).transfer(self, A)
        return [A[:, k] for k in range(self.size)]

    def pdf(self, x):
        '''
        Compute the probability density function (pdf) of each RandomVariable
        in self at points x, x must have self.ndim() columns.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the pdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the pdf at points x.

        '''
        if isinstance(x, list):
            x = np.hstack([np.reshape(y, [-1, 1]) for y in x])

        x = np.atleast_2d(x)

        p = np.zeros(x.shape)
        for i in range(self.size):
            p[:, i] = np.ravel(self.random_variables[i].pdf(x[:, i]))
        p = np.prod(p, axis=1)

        if not isinstance(self.copula, tensap.IndependentCopula):
            u = np.zeros(x.shape)
            for i in range(self.size):
                u[:, i] = np.ravel(self.random_variables[i].cdf(x[:, i]))
            p = p * self.copula.pdf(u)

        return p

    def cdf(self, x):
        '''
        Compute the cumulative distribution function (cdf) at points x, x must
        have self.ndim() columns.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the cdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the cdf at points x.

        '''
        if isinstance(x, list):
            x = np.hstack([np.reshape(y, [-1, 1]) for y in x])

        x = np.atleast_2d(x)

        u = np.zeros(x.shape)
        for i in range(self.size):
            u[:, i] = np.ravel(self.random_variables[i].cdf(x[:, i]))
        return self.copula.cdf(u)

    def orthonormal_polynomials(self, max_degree=None):
        '''
        Return the max_degree-1 first orthonormal polynomials associated with
        the RandomVector.

        Parameters
        ----------
        max_degree : int, optional
            The maximum degree of the returned polynomials. The default is
            None, choosing the default maximum degree associated with the
            constructor of the polynomials.

        Returns
        -------
        poly : tensap.OrthonormalPolynomials
            The generated orthonormal polynomials.

        '''
        if max_degree is None:
            return [x.orthonormal_polynomials() for x in self.random_variables]

        if np.isscalar(max_degree):
            max_degree = np.full(self.size, max_degree)

        assert len(max_degree) == self.size, 'Wrong size for max_degree.'

        return [x.orthonormal_polynomials(max_degree[ind]) for
                ind, x in enumerate(self.random_variables)]

    def mean(self):
        '''
        Return the mean of the random variable.

        Returns
        -------
        float
            The mean of the random variable.

        '''
        return np.atleast_1d([x.mean() for x in self.random_variables])

    def std(self):
        '''
        Return the standard deviation of the random variable.

        Returns
        -------
        float
            The standard deviation of the random variable.

        '''
        return np.atleast_1d([x.std() for x in self.random_variables])

    def variance(self):
        '''
        Return the variance of the random variable.

        Returns
        -------
        float
            The variance of the random variable.

        '''
        return np.atleast_1d([x.variance() for x in self.random_variables])

    def transfer(self, Y, x):
        '''
        Transfer from the tensap.RandomVector X to the tensap.RandomVector Y,
        at points x.

        Parameters
        ----------
        Y : tensap.RandomVector
            The target RandomVector of the transfer.
        x : list or numpy.ndarray
            The input points.

        Returns
        -------
        y : numpy.ndarray
            The transfered points.

        '''
        if isinstance(x, list):
            x = np.hstack([np.reshape(y, [-1, 1]) for y in x])

        x = np.atleast_2d(x)

        if isinstance(Y, tensap.RandomVariable):
            Y = RandomVector(Y, self.size)

        assert self.size == Y.size, \
            'The two RandomVector must have the same dimension.'

        y = np.zeros((x.shape[0], Y.size))
        for i in range(self.size):
            y[:, i] = np.ravel(self.random_variables[i].transfer(
                Y.random_variables[i], x[:, i]))
        return y

    def transpose(self, perm):
        '''
        Transpose (permute) the components of a random vector.

        Parameters
        ----------
        perm : list or numpy.array
            The permutation indices.

        Returns
        -------
        tensap.RandomVector
            The random vector with transposed (permuted) components.

        '''
        return RandomVector(self.random_variables[perm])

    def random(self, n=1):
        '''
        Generate n random numbers according to the distribution of the
        RandomVector.

        Parameters
        ----------
        n : int
            The number of random numbers generated.

        Returns
        -------
        numpy.ndarray
            The generated numbers.

        '''
        assert isinstance(self.copula, tensap.IndependentCopula), \
            'Not implemented for non IndenpendentCopula.'
        return np.transpose(np.array([x.random(n)
                                      for x in self.random_variables]))



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module multivariate_functions_benchmark

'''

import sys
import numpy as np
sys.path.insert(0, './../../../')
import tensap


def multivariate_functions_benchmark(case, *args):
    '''
    Generate different functions used to benchmark the package.

    Parameters
    ----------
    case : str
        The name of the function. Can be 'borehole', 'ishigami',
        'sin_of_a_sum', 'linear_additive', 'linear_rank_one',
        'quadratic_rank_one', 'canonical_rank_2', 'mixture', 'field',
        'oscillatory', 'product_peak', 'corner_peak', 'gaussian', 'continuous',
        'discontinuous', 'henon_heiles', 'sobol', 'anisotropic', 'polynomial'.
    *args : tuple
        Parameters of the function.

    Raises
    ------
    NotImplementedError
        If the function is not implemented.

    Returns
    -------
    function
        The asked function.
    tensap.RandomVector
        Input random variables.

    '''
    if case == 'borehole':
        X = np.empty(8, dtype=object)
        X[0] = tensap.NormalRandomVariable(0.1, 0.0161812)
        X[1] = tensap.NormalRandomVariable(0, 1)
        X[2] = tensap.UniformRandomVariable(63070, 115600)
        X[3] = tensap.UniformRandomVariable(990, 1110)
        X[4] = tensap.UniformRandomVariable(63.1, 116)
        X[5] = tensap.UniformRandomVariable(700, 820)
        X[6] = tensap.UniformRandomVariable(1120, 1680)
        X[7] = tensap.UniformRandomVariable(9855, 12045)
        X = tensap.RandomVector(X)

        def fun(x):
            return 2 * np.pi * x[:, 2] * (x[:, 3] - x[:, 5]) / \
                (np.log(np.exp(7.71 + 1.0056*x[:, 1]) / x[:, 0]) *
                 (1 + 2 * x[:, 6] * x[:, 2] /
                  np.log(np.exp(7.71 + 1.0056 * x[:, 1]) / x[:, 0]) /
                  x[:, 0]**2 / x[:, 7] + x[:, 2] / x[:, 4]))

    elif case == 'ishigami':
        if len(args) == 0:
            d = 3
        else:
            d = args[0]
        if len(args) <= 1:
            a = 7
        else:
            a = args[1]
        if len(args) <= 2:
            b = 0.1
        else:
            b = args[2]

        X = tensap.RandomVector(tensap.UniformRandomVariable(-np.pi, np.pi), d)

        def fun(x):
            return np.sin(x[:, 0]) + a*np.sin(x[:, 1])**2 + \
                b*x[:, 2]**4*np.sin(x[:, 0])

    elif case == 'sin_of_asum':
        if len(args) == 0:
            d = 3
        else:
            d = args[0]
        if len(args) <= 1:
            def fun(x):
                return np.sin(np.sum(x, 1))
        else:
            def fun(x):
                return np.sin(np.matmul(x, np.reshape(args[1], (-1, 1))))

        X = tensap.RandomVector(tensap.UniformRandomVariable(), d)

    elif case == 'linear_additive':
        if len(args) == 0:
            d = 3
        else:
            d = args[0]
        if len(args) <= 1:
            def fun(x):
                return np.sum(x, 1)
        else:
            def fun(x):
                return np.matmul(x, np.reshape(args[1], (-1, 1)))

        X = tensap.RandomVector(tensap.UniformRandomVariable(), d)

    elif case == 'linear_rank_one':
        if len(args) == 0:
            d = 3
        else:
            d = args[0]
        if len(args) <= 1:
            w = np.zeros((1, d))
        else:
            w = np.reshape(args[1], (1, -1))

        X = tensap.RandomVector(tensap.UniformRandomVariable(), d)

        def fun(x):
            return np.prod(np.tile(w, (np.shape(x)[0], 1)) + 2*x, 1)

    elif case == 'quadratic_rank_one':
        d = args[0]
        c = args[1]

        X = tensap.RandomVector(tensap.UniformRandomVariable(), d)

        def fun(x):
            return np.prod(np.tile(c[0, :], (np.shape(x)[0], 1)) +
                           np.tile(c[1, :], (np.shape(x)[0], 1))*x +
                           np.tile(c[2, :], (np.shape(x)[0], 1))*x**2, 1)

    elif case == 'canonical_rank_2':
        if len(args) == 0:
            d = 3
        else:
            d = args[0]

        X = tensap.RandomVector(tensap.UniformRandomVariable(), d)

        def fun(x):
            return x[:, 0] * x[:, 1] * x[:, 2] + x[:, 0]**2 + x[:, 1]

    elif case == 'mixture':
        if len(args) == 0:
            d = 6
        else:
            d = args[0]

        X = tensap.RandomVector(tensap.UniformRandomVariable(), d)

        def fun(x):
            return np.sin(x[:, 0] + x[:, 3])*np.exp(x[:, 4])*x[:, 5] + \
                np.sin(x[:, 2]*x[:, 3])*x[:, 5]

    elif case == 'field':
        if len(args) == 0:
            d = 6
        else:
            d = args[0]

        X = tensap.RandomVector(tensap.UniformRandomVariable(0, 1), d)

        def fun(x):
            return 1 + np.cos(x[:, 0])*x[:, 1] + np.sin(x[:, 0])*x[:, 2] + \
                np.exp(x[:, 0])*x[:, 3] + 1/(x[:, 0]+1)*x[:, 4] + \
                1/(2*x[:, 0]+3)*x[:, 5]

    elif case in ('oscillatory', 'product_peak', 'corner_peak', 'gaussian',
                  'continuous', 'discontinuous'):
        if len(args) == 0:
            d = 10
        else:
            d = args[0]

        w = np.random.rand(1, d)
        c = np.random.rand(1, d)

        X = tensap.RandomVector(tensap.UniformRandomVariable(0, 1), d)

        if case == 'oscillatory':
            def fun(x):
                return np.cos(w[0]*2*np.pi + np.matmul(x, np.transpose(c)))
        elif case == 'product_peak':
            def fun(x):
                return 1/np.prod(np.tile(x**(-2), (np.shape(x)[0], 1)) +
                                 (x + np.tile(w, (np.shape(x)[0], 1)))**2, 1)
        elif case == 'corner_peak':
            b = 185
            e = 2
            c = c*b/d**e/np.sum(c)

            def fun(x):
                return (1+np.matmul(x, np.transpose(c)))**(-d-1)
        elif case == 'gaussian':
            def fun(x):
                return np.matmul(np.exp(
                    -(x - np.tile(w, (np.shape(x)[0], 1)))**2,
                    np.transpose(c)**2))
        elif case == 'continuous':
            def fun(x):
                return np.matmul(np.exp(
                    -np.abs(x - np.tile(w, (np.shape(x)[0], 1))),
                    np.transpose(c)**2))
        elif case == 'discontinuous':
            raise NotImplementedError('Function not implemented.')

    elif case == 'henon_heiles':
        if len(args) == 0:
            d = 3
        else:
            d = args[0]

        X = tensap.RandomVector(tensap.NormalRandomVariable(), d)

        def fun(x):
            return 0.5*np.sum(x**2, 1) + \
                0.2*np.sum(x[:, :-1]*x[:, 1:]**2-x[:, :-1]**3, 1) + \
                0.2**2/16*np.sum((x[:, :-1]**2+x[:, 1:]**2)**2, 1)

    elif case == 'sobol':
        if len(args) == 0:
            d = 8
            a = [1, 2, 5, 10, 20, 50, 100, 500]
        else:
            d = args[0]
            if len(args) <= 1:
                a = 2**np.arange(0, d)
            else:
                a = args[1]

        X = tensap.RandomVector(tensap.UniformRandomVariable(0, 1), d)

        def fun(x):
            return np.prod((np.abs(4*x-2)+np.tile(a, (np.shape(x)[0], 1))) /
                           np.tile(1+a, (np.shape(x)[0], 1)), 1)

    elif case == 'anisotropic':
        if len(args) == 0:
            d = 16
        else:
            d = args[0]

        X = tensap.RandomVector(tensap.UniformRandomVariable(0, 1), d)

        def fun(x):
            return x[:, 2]*np.sin(x[:, 3] + x[:, 15])

    elif case == 'polynomial':
        if len(args) == 0:
            d = 16
        else:
            d = args[0]
        if len(args) <= 1:
            q = 2
        else:
            q = args[1]

        X = tensap.RandomVector(tensap.UniformRandomVariable(0, 1), d)

        def fun(x):
            return 1/(2**np.shape(x)[1])*np.prod(3*x**q+1, 1)

    else:
        raise NotImplementedError('Bad function name.')

    return fun, X



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module tensorized_function.

'''

import numpy as np
import tensap


class TensorizedFunction(tensap.Function):
    '''
    Class TensorizedFunction.

    Attributes
    ----------
    fun : function or tensap.Function
        A function of (d+1)*dim variables.
    tens : tensap.Tensorizer
        A tensap.Tensorizer object.

    '''

    def __init__(self, fun, tens=None):
        '''
        Constructor for the class TensorizedFunction.

        Function g(x1, ..., xdim) identified with a function f(z) of
        (d+1)*dim variables using the Tensorizer tens (base b, resolution d,
        dimension dim).

        For a univariate function g(x) (dim=1),
        g(x) = f(i_1, ..., i_d, y) with y in [0,1] and i_k in {0, ..., b-1}
        where x = (i + y)b^(-d) with i in {0, ..., b^d-1} having the following
        representation in base b:
        i = sum_{k=1}^d i_k b^(d-k) in [0,b^d-1].

        For a bivariate function g(x1, x2) (dim=2)
            - if t.ordering_type == 1 then
            g(x) = f(i_1, ..., i_d, j_1, ...., j_d, y1, y2) with yk in [0,1]
            and x1 = (i + y1)b^(-d), x2 = (j + y2)b^(-d)
            - if t.ordering_type == 2 then
            g(x) = f(i_1, j_1, ..., i_d, j_d, y1, y2).

        Parameters
        ----------
        fun : function or tensap.Function
            A function of (d+1)*dim variables.
        tens : tensap.Tensorizer, optional
            A tensap.Tensorizer object. The default is None, raising an error.

        Returns
        -------
        None.

        '''
        assert tens is not None, 'Must provide a Tensorizer.'

        tensap.Function.__init__(self)

        self.fun = fun
        self.tens = tens
        self.dim = tens.dim

    def eval(self, z):
        z = np.array(z)
        if z.ndim == 1:
            z = np.reshape(z, [-1, 1])
        if z.shape[1] == self.tens.dim:
            z = self.tens.map(z)
        return self.fun(z)

    def domain(self):
        '''
        Return the domain of the function.

        Returns
        -------
        sup : numpy.ndarray
            The domain of the function.

        '''
        return self.tens.X.support()



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on multivariate functions, Tensor Grids, Projection.

'''

import matplotlib.pyplot as plt
import numpy as np
import tensap

# %% Scalar-valued UserDefinedFunction
d = 3
fun = tensap.UserDefinedFunction('x0+x1+x2**4', d)
fun.evaluation_at_multiple_points = True
x = np.random.rand(4, d)
print('fun.eval(x) = \n%s\n' % fun.eval(x))
print('fun(x) =      \n%s\n' % fun(x))

# %% Vector-valued UserDefinedFunction
f = tensap.UserDefinedFunction('[x0, 100*x1]', d, 2)
f.evaluation_at_multiple_points = False
print('With f.evaluation_at_multiple_points == False, f(x) = \n%s\n' % f(x))
f.evaluation_at_multiple_points = True
print('With f.evaluation_at_multiple_points == True, f(x) =  \n%s\n' % f(x))

f = tensap.UserDefinedFunction('[x1,100*x2]', d, [1, 2])
f.evaluation_at_multiple_points = False
print('With f.evaluation_at_multiple_points == False, f(x) = \n%s\n' % f(x))
f.evaluation_at_multiple_points = True
print('With f.evaluation_at_multiple_points == True, f(x) =  \n%s\n' % f(x))

# %% Evaluation on a FullTensorGrid
g = tensap.FullTensorGrid(np.linspace(-1, 1, 50), d)
fx = fun.eval_on_tensor_grid(g)
g.plot_grid(marker='x')

# %% Evaluation on a SparseTensorGrid
I = tensap.MultiIndices.with_bounded_norm(d, 1, 5)
g = tensap.SparseTensorGrid(np.linspace(0, 1, 6), I, d)
fx = fun.eval_on_tensor_grid(g)
g.plot_grid(marker='x')

# %% Functional Bases
h = tensap.PolynomialFunctionalBasis(tensap.CanonicalPolynomials(), range(5))
H = tensap.FunctionalBases.duplicate(h, d)
grid = tensap.FullTensorGrid(np.linspace(-1, 1, 10), d)
x = grid.array()
Hx = H.eval(x)

# %% Sparse tensor functional basis
d = 2
p = 1
m = 4
h = tensap.PolynomialFunctionalBasis(tensap.CanonicalPolynomials(), range(5))
H = tensap.FunctionalBases.duplicate(h, d)
I = tensap.MultiIndices.with_bounded_norm(d, p, m)
Psi = tensap.SparseTensorProductFunctionalBasis(H, I)

print('Multi-indices: \n%s\n' % I.array)

finegrid = tensap.FullTensorGrid(np.arange(-1, 1.1, 0.1), d)
x = finegrid.array()
Psix = Psi.eval(x)

for i in range(I.cardinal()):
    finegrid.plot(Psix[:, i])
    plt.title(I.array[i, :])

# %% Projection on polynomial space through quadrature
d = 3
p = 3
fun = tensap.UserDefinedFunction('x0+x1**2+x2**3', d)
fun.evaluation_at_multiple_points = True
v = tensap.NormalRandomVariable()
X = tensap.RandomVector(v, d)
I = v.gauss_integration_rule(5).tensorize(d)
u = fun.eval_on_tensor_grid(I.points)

h = tensap.PolynomialFunctionalBasis(tensap.HermitePolynomials(), range(p+1))
H = tensap.FunctionalBases.duplicate(h, d)
H = tensap.FullTensorProductFunctionalBasis(H)

f, _ = H.projection(fun, I)

N_test = 100
x_test = X.random(N_test)
err_test = np.linalg.norm(f(x_test) - fun(x_test))/np.linalg.norm(fun(x_test))
print('Test error = %2.5e\n' % err_test)
